import ujson
import psycopg2
import psycopg2.pool
import psycopg2.extras
from .. import eventlog
from .. import path_helpers as ph
from ..com import fs

psycopg2.extras.register_default_json(globally=True, loads=ujson.loads)
psycopg2.extras.register_default_jsonb(globally=True, loads=ujson.loads)


class DbError(Exception):
    pass


class DbHintError(DbError):
    def __init__(self, message, hint):
        """
        :type hint: str
        """
        DbError.__init__(self, message)
        self.hint = hint


class DbQueryError(DbHintError):
    pass


class DbNotFoundError(DbHintError):
    pass


class DbInternalError(DbHintError):
    pass


_x_map = {
    '400': DbQueryError,
    '404': DbNotFoundError,
    '500': DbInternalError
}


def trap_hint_error(x):
    """
    :type x: psycopg2.Error
    :raise DbHintError:
    """
    error = str(x.pgerror)
    for line in error.splitlines():
        if line.startswith('HINT'):
            hint = line.split(' ')[-1]
            raise _x_map.get(hint, DbHintError)(x.message, hint)


class DbJsonArg(psycopg2.extras.Json):
    def dumps(self, jnode):
        """
        :type jnode: dict | list
        """
        return ujson.dumps(jnode)


class DbArgs:
    def __init__(self, db_user, db_pass, db_name, db_host='localhost', db_port=5433):
        """
        :type db_user: str
        :type db_pass: str
        :type db_name: str
        :type db_port: int
        :type db_host: str
        """
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_host = db_host
        self.db_port = db_port

    @staticmethod
    def deserialize(db_jnode):
        """
        :type db_jnode: dict
        :rtype: DbArgs

        :raise: KeyError
        """
        db_name = db_jnode['name']
        db_user = db_jnode['user']
        db_pass = db_jnode['pass']
        db_host = db_jnode['host']
        db_port = db_jnode['port']
        return DbArgs(db_user, db_pass, db_name, db_host, db_port)

    @staticmethod
    def deserialize2(filepath=ph.DB_JSON_FILEPATH):
        """
        :type filepath: str
        :rtype: DbArgs

        :raise: KeyError
        :raise: ValueError
        """
        db_jnode = fs.load_json(filepath)
        if db_jnode:
            db_args = DbArgs.deserialize(db_jnode['db'])
            return db_args

        eventlog.log_e('fail: invalid {}'.format(filepath))
        raise ValueError


def with_db(db_args, handler):
    """
    :type db_args:DbArgs 
    :type handler: (Any) -> Any
    :rtype: Any
    :raises: DbError
    """
    try:
        db_link = psycopg2.connect(
            database=db_args.db_name, user=db_args.db_user, password=db_args.db_pass,
            host=db_args.db_host, port=db_args.db_port)
        try:
            return handler(db_link)
        finally:
            db_link.close()

    except psycopg2.Error as x:
        eventlog.log_e('fail: {}'.format(x))
        raise DbError
