import os
import apt
import hashlib


from .types import AboutServiceArgs
from ..com import fs
from ..com import net
from ..model.types import Origin, Location
from ..service.gpsd import IGpsdService
from ..path_helpers import ABOUT_JSON_FILEPATH
from .. import eventlog

SW_DIRPATH = '/opt/edge/share/dpkg/info'


class AboutError(RuntimeError):
    pass


class IAboutService:
    def __init__(self):
        pass

    def get_origin(self, ts):
        """
        :type ts: long
        :rtype: Origin
        """
        raise NotImplementedError

    def get_hw_serial(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    def get_hw_vendor(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    def get_hw_model(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def hw_info(self):
        """
        :rtype: list
        """
        return '', '', ''

    @property
    def sw_info(self):
        """
        :rtype: list
        """
        raise NotImplementedError

    @property
    def sw_name(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def base_url(self):
        """
        :rtype: str | None
        """
        raise NotImplementedError


def _attach_files_md5(sw_files_jnode):
    """
    :param sw_files_jnode: *.sw/#/files
    :type sw_files_jnode: list[dict]
    :rtype: dict
    """
    blocksize = 65536

    for file_jnode in sw_files_jnode:
        hasher = hashlib.md5()
        file_path = file_jnode['path']
        with open(file_path, 'rb') as f:
            buf = f.read(blocksize)
            while len(buf) > 0:
                hasher.update(buf)
                buf = f.read(blocksize)

        file_jnode['md5'] = hasher.hexdigest()

    return sw_files_jnode


def get_hw_version():
    """
    :rtype: dict | None
    """
    about_jnode = fs.load_json(ABOUT_JSON_FILEPATH) or dict()
    return about_jnode.get('hw')


def get_sw_name():
    """
    :rtype: dict | None
    """
    about_jnode = fs.load_json(ABOUT_JSON_FILEPATH) or dict()
    return about_jnode.get('sw-name')


def get_sw_lang():
    """
    :rtype: list(str) | None
    """
    about_jnode = fs.load_json(ABOUT_JSON_FILEPATH) or dict()
    return about_jnode.get('sw-lang')


def get_sw_version(sw_files_dir):
    """
    :param sw_files_dir: /opt/edge/share/dpkg/info/
    :type sw_files_dir: str
    :rtype: list[dict]
    """
    ver = None
    answer = []

    about_jnode = fs.load_json(ABOUT_JSON_FILEPATH) or dict()
    if about_jnode:
        ver = about_jnode.get('sw-ver')

    if ver:
        answer.append({
            'package': ['VIT', "edge", ver]
        })

    extra_sw_files = ['edge.sw'] if not ver else []

    try:
        all_files = os.listdir(sw_files_dir)
    except os.error as x:
        eventlog.log_e('fail: os.listdir <%s>' % x)
        all_files = []

    sw_files = extra_sw_files + filter(lambda name: name.endswith('.sw'), all_files)
    with apt.cache.Cache() as apt_cache:
        for sw_name in sw_files:
            package_name = sw_name.split('.')[0]
            # package_info = apt_cache.get(package_name)
            package_info = None
            try:
                package_info = apt_cache[package_name]
            except KeyError:
                pass

            if not package_info:
                continue

            version_info = package_info.installed
            if not version_info:
                continue

            sw_path = fs.join_path(sw_files_dir, sw_name)
            sw_jnode = fs.load_json(sw_path)
            sw_files_jnode = sw_jnode.get('files', []) if sw_jnode else []

            if sw_files_jnode:
                answer.append({
                    'package': ['VIT', package_name, version_info.version],
                    'files': _attach_files_md5(sw_files_jnode)
                })
            else:
                answer.append({
                    'package': ['VIT', package_name, version_info.version]
                })

    return answer


class _Service(IAboutService):
    def __init__(self, args, gpsd_service):
        """
        :type args: AboutServiceArgs
        :type gpsd_service: IGpsdService
        """
        IAboutService.__init__(self)
        self._gpsd = gpsd_service

        def get_hw_info():
            """
            :rtype: (str, str, str)
            """
            jnode = fs.load_json(args.server_about_path) or dict()
            hw_jnode = jnode.get('hw') or list()
            return hw_jnode if len(hw_jnode) == 3 else ['', '', '']

        def mk_origin_factory():
            jnode = fs.load_json(args.origin_about_path) or dict()
            o_jnode = jnode.get('origin') or dict()
            l_jnode = jnode.get('location') or dict()

            uuid = o_jnode['uuid']
            ctime = o_jnode['time']
            place = l_jnode.get('place') or ''

            return lambda geolocation: Origin(uuid, ctime, Location(place=place), geolocation)

        self._hw_info = get_hw_info()
        self._mk_origin = mk_origin_factory()
        self._sw_info = get_sw_version(SW_DIRPATH)
        self._sw_name = get_sw_name() or 'EDGE'
        self._base_url = None

    @property
    def hw_info(self):
        """
        :rtype: list
        """
        return self._hw_info

    @property
    def sw_info(self):
        """
        :rtype: list
        """
        return self._sw_info

    @property
    def sw_name(self):
        """
        :rtype: str
        """
        return self._sw_name

    @property
    def base_url(self):
        """
        :rtype: str
        """
        if self._base_url is None:
            try:
                endpoint = net.get_http_api_endpoint()
                self._base_url = '{}/api/v2/origins/event/dist'.format(endpoint)
            except net.NetError:
                eventlog.log_w('fail: define <base_url>')

        return self._base_url

    def get_origin(self, ts):
        """
        :type ts: long
        :rtype: Origin
        """
        geolocation = self._gpsd.get_geodata(ts)
        return self._mk_origin(geolocation)

    def get_hw_vendor(self):
        """
        :rtype: str
        """
        return self._hw_info[0]

    def get_hw_model(self):
        """
        :rtype: str
        """
        return self._hw_info[1]

    def get_hw_serial(self):
        """
        :rtype: str
        """
        return self._hw_info[2]


def init_service(args, gpsd_service):
    """
    :type args: AboutServiceArgs
    :type gpsd_service: IGpsdService
    :rtype: IAboutService
    """
    return _Service(args, gpsd_service)


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: IAboutService | _Service
    """
    pass
