from twisted.internet.protocol import ReconnectingClientFactory, connectionDone
from twisted.protocols.basic import LineOnlyReceiver

import sortedcontainers
import gps.misc
import ujson

from edge import eventlog

from .types import GpsdServiceArgs, GpsdCounters
from ..model.geolocation import Geolocation

_TPV_MODE_NONE = 0
_TPV_MODE_NO_FIX = 1
_TPV_MODE_2D_FIX = 2
_TPV_MODE_3D_FIX = 3


# noinspection PyMethodMayBeStatic
class IGpsdService:
    def __init__(self):
        pass

    def get_geodata(self, ts):
        """
        :type ts: int
        :rtype: Geolocation | None
        """
        raise NotImplementedError


def init_service(twisted_reactor, args, counters=None):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
    :type args: GpsdServiceArgs
    :type counters: GpsdCounters | None
    :rtype: IGpsdService
    """
    if args.provider == args.PROVIDER_STATIC:
        gpsd = _StubGpsdService(args.geodata)
    else:
        gpsd = _GpsdService(twisted_reactor, args, counters)

    return gpsd


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: IGpsdService | _GpsdService
    """
    pass


class _StubGpsdService(IGpsdService):
    def __init__(self, geodata):
        """
        :type geodata: Geolocation
        """
        IGpsdService.__init__(self)

        # http://catb.org/gpsd/gpsd_json.html
        self._geodata = geodata

    def get_geodata(self, ts):
        """
        :type ts: int
        :rtype: Geolocation
        """
        return self._geodata


class _GeodataStore:
    def __init__(self, max_entry_count, max_age_delta_usec, max_lookup_dist_usec):
        """
        :type max_entry_count: int
        :type max_age_delta_usec: int
        :type max_lookup_dist_usec: int
        """
        self._max_entry_count = max_entry_count
        self._max_age_delta_usec = max_age_delta_usec
        self._max_lookup_dist_usec = max_lookup_dist_usec

        self._ts_to_geodata_map = sortedcontainers.SortedDict()

    def add_entry(self, ts, geodata):
        """
        :type ts: int
        :type geodata: Geolocation
        """
        self._ts_to_geodata_map[ts] = geodata

        while len(self._ts_to_geodata_map) > self._max_entry_count:
            self._ts_to_geodata_map.popitem(last=False)

        while (self._ts_to_geodata_map.peekitem()[0] - self._ts_to_geodata_map.peekitem(0)[0]) \
                > self._max_age_delta_usec:
            self._ts_to_geodata_map.popitem(last=False)

    def lookup_nearest(self, lookup_ts):
        """
        :type lookup_ts: int
        :rtype: Geolocation | None
        """
        entry_count = len(self._ts_to_geodata_map)
        if not entry_count:
            return None

        # insert lookup_ts
        bisect_index = self._ts_to_geodata_map.bisect_right(lookup_ts)
        if 0 < bisect_index < entry_count:
            # exact-match with lookup_ts
            less_equal_ts_dist = lookup_ts - self._ts_to_geodata_map.iloc[bisect_index - 1]
            if less_equal_ts_dist == 0:
                return self._ts_to_geodata_map.peekitem(bisect_index - 1)[1]

            # greater_ts is better
            greater_ts_dist = self._ts_to_geodata_map.iloc[bisect_index] - lookup_ts
            if less_equal_ts_dist < greater_ts_dist and less_equal_ts_dist <= self._max_lookup_dist_usec:
                return self._ts_to_geodata_map.peekitem(bisect_index - 1)[1]

            # less_equal_ts is better
            return self._ts_to_geodata_map.peekitem(bisect_index)[1] \
                if greater_ts_dist <= self._max_lookup_dist_usec else None

        # push-back lookup_ts
        if bisect_index == entry_count:
            less_equal_ts_dist = lookup_ts - self._ts_to_geodata_map.iloc[entry_count - 1]
            return self._ts_to_geodata_map.peekitem(bisect_index - 1)[1] \
                if less_equal_ts_dist <= self._max_lookup_dist_usec else None

        # push-front lookup_ts
        greater_ts_dist = self._ts_to_geodata_map.iloc[bisect_index] - lookup_ts
        return self._ts_to_geodata_map.peekitem(bisect_index)[1] \
            if greater_ts_dist <= self._max_lookup_dist_usec else None

    def get_count(self):
        return len(self._ts_to_geodata_map)


class _GpsdProtocol(LineOnlyReceiver):
    delimiter = b'\n'

    def __init__(self, geodata_store, stat_counters=None):
        """
        :type geodata_store: _GeodataStore
        :type stat_counters: GpsdCounters | None
        """
        self._geodata_store = geodata_store
        self._stat_counters = stat_counters

    def connectionMade(self):
        # noinspection PyBroadException
        try:
            self.sendLine('?WATCH={"enable":true,"json":true}')
        except Exception:
            eventlog.log_x()

    # Called when an established connection is lost.
    def connectionLost(self, reason=connectionDone):
        pass

    def lineReceived(self, raw_reply):
        try:
            self._handle(raw_reply)

        except ValueError:
            eventlog.log_x()

    def _handle(self, raw_reply):
        """
        :type raw_reply: str
        """
        reply_jnode = ujson.loads(raw_reply)
        if reply_jnode['class'] == 'TPV':
            # Time/date stamp in ISO8601 format, UTC.
            # May have a fractional part of up to .001sec precision. May be absent if mode is not 2 or 3.
            raw_tpv_time = reply_jnode.get('time')
            tpv_time_usec = None
            tpv_time_sec = None
            if raw_tpv_time:
                tpv_iso_time = gps.isotime(raw_tpv_time.encode("ascii"))
                tpv_time_usec = int(tpv_iso_time * 1e6)
                tpv_time_sec = int(tpv_iso_time)

            # Latitude in degrees: +/- signifies North/South. Present when mode is 2 or 3.
            tpv_lat = reply_jnode.get('lat')

            # Longitude in degrees: +/- signifies East/West. Present when mode is 2 or 3.
            tpv_lon = reply_jnode.get('lon')

            # Altitude in meters. Present if mode is 3.
            tpv_alt = reply_jnode.get('alt')

            # Speed over ground, meters per second.
            tpv_speed = reply_jnode.get('speed')

            # TODO: implement stub stat counters
            if self._stat_counters:
                self._stat_counters.geodata.inc(tpv_time_sec, tpv_lat, tpv_lon, tpv_alt)

            # NMEA mode: %d, 0=no mode value yet seen, 1=no fix, 2=2D, 3=3D.
            tpv_mode = reply_jnode['mode']
            if self._stat_counters:
                self._stat_counters.inc_mode(tpv_mode)

            if tpv_time_usec is not None and tpv_mode in [_TPV_MODE_2D_FIX, _TPV_MODE_3D_FIX]:
                self._geodata_store.add_entry(tpv_time_usec, Geolocation(tpv_lat, tpv_lon, tpv_alt, tpv_speed))
                if self._stat_counters:
                    self._stat_counters.inc_store_count(self._geodata_store.get_count())


class _GpsdClientFactory(ReconnectingClientFactory):
    initialDelay = 1
    maxDelay = 15
    factor = 1.6180339887498948

    def __init__(self, geodata_store, stat_counters=None):
        """
        :type geodata_store: _GeodataStore
        :type stat_counters: GpsdCounters | None
        """
        self._geodata_store = geodata_store
        self._stat_counters = stat_counters

    def buildProtocol(self, addr):
        self.resetDelay()
        protocol = _GpsdProtocol(self._geodata_store, self._stat_counters)
        protocol.factory = self
        return protocol

    def clientConnectionFailed(self, connector, reason):
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, unused_reason):
        ReconnectingClientFactory.clientConnectionLost(self, connector, unused_reason)


class _BaseGpsdService(IGpsdService):
    def __init__(self, twisted_reactor, endpoint, geodata_store, stat_counters=None, azimuth=None):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
        :type endpoint: (str, int)
        :type geodata_store: _GeodataStore
        :type stat_counters: GpsdCounters | None
        :type azimuth: float | None
        """
        IGpsdService.__init__(self)

        self._azimuth = azimuth
        self._geodata_store = geodata_store

        gpsd_host, gpsd_port = endpoint
        client_factory = _GpsdClientFactory(self._geodata_store, stat_counters)
        twisted_reactor.connectTCP(gpsd_host, gpsd_port, client_factory)

    def get_geodata(self, ts):
        """
        :type ts: int
        :rtype: Geolocation | None
        """
        geo = self._geodata_store.lookup_nearest(ts)
        if geo is not None:
            geo.azimuth = self._azimuth
        return geo


class _GpsdService(_BaseGpsdService):
    def __init__(self, twisted_reactor, args, stat_counters=None):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
        :type args: GpsdServiceArgs
        :type stat_counters: GpsdCounters | None
        """
        geodata_store = _GeodataStore(args.max_entry_count, args.max_age_delta_usec, args.max_lookup_dist_usec)
        _BaseGpsdService.__init__(self, twisted_reactor, args.endpoint, geodata_store, stat_counters,
                                  azimuth=args.azimuth)
