import datetime
import iso8601

from .types import TzServiceArgs


def datetime_to_isostr(dt, microseconds=False):
    """
    :type dt: datetime.datetime
    :type microseconds: bool
    :rtype: str
    """
    tz_suffix = dt.strftime('%z')
    if tz_suffix == '+0000':
        tz_suffix = 'Z'

    precision = 6 if microseconds else 3
    precision += 1
    return dt.strftime('%Y-%m-%dT%H:%M:%S') + dt.strftime('.%f')[:precision] + tz_suffix


def isostr_to_datetime(isodate):
    """
    :type isodate: str
    :rtype: datetime.datetime
    """
    return iso8601.parse_date(isodate)


class ITzService:
    def __init__(self):
        pass

    def get_datetime_isostr(self, ts):
        """
        :type ts: ts
        :rtype: str
        """
        raise NotImplementedError

    def get_datetime(self, ts):
        """
        :type ts: ts
        :rtype: datetime.datetime
        """
        raise NotImplementedError

    def get_datetime_now_isostr(self):
        """
        :rtype: str
        """
        raise NotImplementedError


def init_service(args):
    """
    :type args: TzServiceArgs
    :rtype: ITzService
    """
    return _TzService(args)


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: ITzService | _TzService
    """
    pass


class _TzService(ITzService):
    def __init__(self, args):
        """
        :type args: TzServiceArgs
        """
        ITzService.__init__(self)
        self._tz = args.tz_info

    def get_datetime_isostr(self, ts):
        """
        :param ts: time in seconds since the epoch
        :type ts: float | long | int
        :rtype: str
        """
        # 2018-01-23T15:49:47.124+0400
        dt = self.get_datetime(ts)
        return datetime_to_isostr(dt)

    def get_datetime(self, ts):
        """
        :param ts: time in seconds since the epoch
        :type ts: float | long | int
        :rtype: datetime.datetime
        """
        return datetime.datetime.fromtimestamp(float(ts), tz=self._tz)

    def get_datetime_now_isostr(self):
        """
        :rtype: str
        """
        dt = datetime.datetime.now(self._tz)
        return datetime_to_isostr(dt)
